(*********************************************************************)
(*                         Diy                                       *)
(*                                                                   *)
(*   Jade Alglave, Luc Maranget INRIA Paris-Rocquencourt, France.    *)
(*                                                                   *)
(*  Copyright 2010 Institut National de Recherche en Informatique et *)
(*  en Automatique. All rights reserved. This file is distributed    *)
(*  under the terms of the Lesser GNU General Public License.        *)
(*********************************************************************)

open Misc
open Printf

(* Configuration *)
let arch = ref PPC
let norm = ref false
let () = Config.nprocs := 1000
let () = Config.numeric := true

let opts =
  Config.common_specs @
  ("-num", Arg.Bool (fun b -> Config.numeric := b),
   sprintf "<bool> use numeric names, default %b" !Config.numeric)::
  ("-norm",Arg.Set norm," find a mormalised name for me")::
  Util.arch_opt (fun a -> arch := a) (pp_arch !arch)



module type Co = sig
  include DumpAll.Config
  val norm : bool
end

module Make(Co:Co) (C:XXXCompile.S) =
  struct
    module M = Top.Make(Co)(C)

    let dump_stdout es =
      let t = M.make_test "A" es in
      M.dump_test_channel stdout t ;
      None

    let litmus name = sprintf "%s.litmus" name

    let dump_file name es =
      if !verbose > 0 then eprintf "Test name: %s\n" name ;
      let t = M.make_test name es in
      let fname = litmus name in
      Misc.output_protect
        (fun chan -> M.dump_test_channel chan t; Some fname)
        fname

    let dump =
      if Co.norm then
        let module Namer =
          Namer.Make(C.A)(C.E) in
        let module Normer = Normaliser.Make(C.E) in
        fun name es ->
          let base,es = Normer.normalise_family es in
          let name = Namer.mk_name base es in
          dump_file name es
      else
        fun name  -> match name with
        | None -> dump_stdout
        | Some name -> dump_file name

    let parse_line s =
      try
        let r = String.index s ':' in
        let name  = String.sub s 0 r
        and es = String.sub s (r+1) (String.length s - (r+1)) in
        let es = LexUtil.split es in
        let es = C.R.parse_relaxs es in
        let es =
          List.fold_right
            (fun r k -> C.R.edges_of r @ k)
            es [] in
        name,es
      with
      | Not_found | Invalid_argument _ ->
          Warn.fatal "bad line: %s" s

(********)
    let do_zyva name pp_rs =
      try
        let pp_rs = List.map LexUtil.split pp_rs in
        let pp_rs = List.concat pp_rs in
        let rs = List.map C.R.parse_relax pp_rs in
        let es =
          List.fold_right
            (fun r k -> C.R.edges_of r @ k)
            rs [] in
        match es with
        | [] ->
            let dump_names =
              Co.norm ||
              (match Co.family with Some _ -> true | None -> false) in
            let module DumpConfig = struct
              include Co
            end in
            let module D = DumpAll.Make(DumpConfig)(C) in
            let gen kont =
              let rec do_rec k0 =
                let k = 
                  try
                    let line = read_line () in
                    try
                      let name,es = parse_line line in
                      let mk_name =
                        if dump_names then  D.no_name
                        else (fun _ -> Some name) in
                      Some (kont es D.no_info mk_name k0)
                    with
                    | Fatal msg | UserError msg ->
                        Warn.fatal "%s on line '%s'" msg line
                  with
                  | End_of_file -> None in
                match k with
                | None -> k0
                | Some k -> do_rec k in
              do_rec in
            D.all gen
        | _ -> ignore (dump name es)
            with Fatal msg ->
              eprintf "Fatal error: %s\n" msg ;
              exit 2

    let zyva = do_zyva Co.family

  end

let pp_es = ref []

let () =
  Util.parse_cmdline
    opts
    (fun x -> pp_es := x :: !pp_es)

let pp_es = List.rev !pp_es

let () =
  let module V = SymbConstant in
  let module Co = struct
(* Dump all *)
    let verbose = !Misc.verbose
    let hout = match !Config.hout with
    | None -> Hint.none
    | Some n -> Hint.open_out n
    let family = !Config.name
    let canonical_only = !Config.canonical_only
    let fmt = !Config.fmt
    let no = match !Config.no with
    | None -> []
    | Some fname -> Config.read_no fname
    let unicond = !Config.unicond
    let tarfile = !Config.tarfile
    let addnum = !Config.addnum
    let numeric = !Config.numeric
    let norm = !norm
  end in
  let module Build = Make(Co) in
  let module C = struct
    let list_edges = !Config.list_edges
    let coherence_decreasing = !Config.coherence_decreasing
    let same_loc = !Config.same_loc || !Config.unicond 
    let sta = !Config.sta
    let unrollatomic = !Config.unrollatomic
  end in
  (match !arch with
  | X86 ->
      let module M = Build(X86Compile.Make(V)(C)) in
      M.zyva
  | PPC ->
      let module M = Build(PPCCompile.Make(V)(C)(PPCArch.Config)) in
      M.zyva
  | ARM ->
      let module M = Build(ARMCompile.Make(V)(C)) in
      M.zyva)
    pp_es

