(*********************************************************************)
(*                          Litmus                                   *)
(*                                                                   *)
(*        Luc Maranget, INRIA Paris-Rocquencourt, France.            *)
(*        Susmit Sarkar, University of Cambridge, UK.                *)
(*                                                                   *)
(*  Copyright 2010 Institut National de Recherche en Informatique et *)
(*  en Automatique and the authors. All rights reserved.             *)
(*  This file is distributed  under the terms of the Lesser GNU      *)
(*  General Public License.                                          *)
(*********************************************************************)

(******************************)
(* Test configuration control *)
(******************************)

(* OS *)
type target_os = Linux | Mac |AIX
(* type of barrier *)
type barrier =
    User | Pthread | NoBarrier | User2 | TimeBase | UserFence
(* how threads are started/joined *)
type launch = Fixed | Changing
type memory = Direct | Indirect
type word = W32 | W64 |WXX
type safer = NoSafe|AllSafe|WriteSafe

type config =
    { gcc_opts : string ;
      barrier : barrier ;
      launch : launch ;
      size_of_test : int ;
      number_of_run : int ;
      memory : memory ;
      safer : safer ;
      os : target_os ;
      word : word ;
      avail : int option ;
      delay : int ;
    }

let default =
   { gcc_opts = "" ;
    barrier = User ;
    launch = Changing ;
    size_of_test = 100000 ;
    number_of_run = 10 ;
    memory = Indirect ;
    safer = WriteSafe ;
    os = Linux ;
    word = WXX ;
    avail = None ;
    delay = 1024 ; }


let default_x86 =
  { default with
    gcc_opts = "-fomit-frame-pointer -O2" ;
    delay = 2048; }

let default_ppc =
  { default with
    gcc_opts = "-O2" ;
    delay = 1024 ; }

let default_arm = default_ppc

let config = ref None


let mod_config = ref (fun cfg -> cfg)

let get_default arch = match arch with
| Misc.X86 -> default_x86
| Misc.PPC -> default_ppc
| Misc.ARM -> default_arm

let get_config arch =
  let cfg = match !config with
  | None -> get_default arch
  | Some cfg -> cfg in
  !mod_config cfg


let replace_config f =
  let g = !mod_config in
  mod_config := (fun cfg -> f (g cfg))

open Printf

let dump_def chan var x =
  fprintf chan "#ifndef %s\n" var ;
  fprintf chan "#define %s %s\n" var x ;
  fprintf chan "#endif\n" ;
  ()

let default_size () = default.size_of_test

let set_size  n = replace_config (fun cfg -> { cfg with size_of_test=n ; })

let get_size arch = (get_config arch).size_of_test

let default_run () = default.number_of_run

let set_run n = replace_config (fun cfg -> { cfg with number_of_run=n; })

let get_run arch = (get_config arch).number_of_run

let dump_bench arch chan =
  let config = get_config arch in
  dump_def chan "SIZE_OF_TEST" (string_of_int config.size_of_test) ;
  dump_def chan "NUMBER_OF_RUN" (string_of_int config.number_of_run) ;
  dump_def chan "AVAIL"
    (match config.avail with
    | None -> "1" | Some n -> sprintf "%i" n) ;
  ()
  

let set_gcc_opts opts =
  replace_config
    (fun cfg -> { cfg with gcc_opts = opts ; })

and get_gcc_opts arch = (get_config arch).gcc_opts

let default_gcc_opts arch = match arch with
| Misc.X86 -> default_x86.gcc_opts
| Misc.PPC -> default_ppc.gcc_opts
| Misc.ARM -> default_arm.gcc_opts

let set_barrier tag =
  let b = match tag with
  | "user" -> User
  | "userfence" -> UserFence
  | "user2" -> User2
  | "pthread" -> Pthread
  | "none" -> NoBarrier
  | "timebase"|"tb" -> TimeBase
  | s -> raise (Arg.Bad ("incorrect barrier tag: "^s)) in
  replace_config (fun cfg -> { cfg with barrier = b })

and get_barrier arch = (get_config arch).barrier

let do_pp_barrier = function
| User -> "user"
| User2 -> "user2"
| UserFence -> "userfence"
| Pthread -> "pthread"
| NoBarrier -> "none"
| TimeBase  -> "timebase"

let pp_barrier arch = do_pp_barrier (get_barrier arch)
let default_barrier () = do_pp_barrier default.barrier

(* Show time of barrier synchronisation *)
let vb = ref false
let get_verbose_barrier () = !vb
let set_verbose_barrier b = vb := b

let xy = ref false
let get_xy () = !xy
let set_xy b = xy := b

(* Contiguous memory *)
let contiguous = ref false
let get_contiguous () = !contiguous
let set_contiguous b = contiguous := b

(* Gnu as syntax *)
let gas = ref None

let set_gas b = gas := Some b
let get_gas () = !gas

(* OS *)

let get_target_os arch = (get_config arch).os

let set_os tag =
  let os = match String.lowercase tag with
  | "linux" -> Linux
  | "mac"|"macos" -> Mac
  | "aix"| "aix5" -> AIX
  | _ -> raise (Arg.Bad ("incorrect os tag: "^tag)) in
  replace_config (fun cfg -> { cfg with os = os; })


let get_word arch = (get_config arch).word

let do_pp_os = function
  | Linux -> "linux"
  | Mac -> "mac"
  | AIX -> "aix"


let pp_os = do_pp_os

let default_os () = do_pp_os default.os

let set_word tag =
  let w = match String.lowercase tag with
  | "w32" -> W32
  | "w64" -> W64
  | _ -> raise (Arg.Bad ("incorrect word tag: "^tag)) in
  replace_config (fun cfg -> { cfg with word = w; })

let get_word arch = (get_config arch).word

let do_pp_ws = function
  | W32 -> "w32"
  | W64 -> "w64"
  | WXX -> "-"

let default_ws () = do_pp_ws default.word

let get_numeric_labels arch = match get_gas () with
| Some b -> b
| None ->
    begin match get_target_os arch with
    | AIX -> false
    | Linux|Mac -> true
    end


(* Launch technique *)
let set_launch tag =
  let l = match tag with
  | "fixed" -> Fixed
  | "changing" -> Changing
  | s -> raise (Arg.Bad ("incorrect launch tag: "^s)) in
  replace_config (fun cfg -> { cfg with launch = l })

and get_launch arch = (get_config arch).launch

let do_pp_launch = function
| Fixed -> "fixed"
| Changing -> "changing"

let pp_launch arch = do_pp_launch (get_launch arch)

let default_launch () = do_pp_launch default.launch

(* Memory *)
let set_memory tag = 
 let l = match tag with
  | "direct" -> Direct
  | "indirect" -> Indirect
  | s -> raise (Arg.Bad ("incorrect mem tag: "^s)) in
  replace_config (fun cfg -> { cfg with memory = l })

and get_memory arch = (get_config arch).memory

let do_pp_memory = function
| Direct -> "direct"
| Indirect -> "indirect"

let pp_memory arch = do_pp_memory (get_memory arch)
let default_memory arch = do_pp_memory default.memory

let set_delay i = replace_config (fun cfg -> { cfg with delay = i })
let get_delay arch = (get_config arch).delay
let default_delay a = (get_default a).delay

(* Stride *)
let default_stride () = None
let stride = ref (default_stride ())
let set_stride i = stride := if i > 0 then Some i else None
let get_stride () = !stride
let pp_stride = function
  | None -> "seq"
  | Some i -> sprintf "%i" i


(* Safer mode *)
let parse_safer tag = match tag with
| "false"|"no" -> NoSafe
| "true"|"all" -> AllSafe
| "write" -> WriteSafe
| _ -> raise (Arg.Bad ("incorrect safer tag: "^tag))

let do_pp_safer = function
| NoSafe -> "no"
| AllSafe -> "all"
| WriteSafe -> "write"

let set_safer s =
  let v = parse_safer s in
  replace_config (fun cfg -> { cfg with safer=v })

let get_safer arch = (get_config arch).safer
let pp_safer arch = do_pp_safer (get_safer arch)
let default_safer () = do_pp_safer default.safer

(* Preload of shared globals by each proc *)
type preload = NoPL|RandomPL|CustomPL
let dpreload = RandomPL
let preload = ref dpreload

let parse_preload = function
  | "false"|"no" -> NoPL
  | "true"|"random" -> RandomPL
  | "custom" -> CustomPL
  | tag -> raise (Arg.Bad ("incorrect preload tag: "^tag))

let pp_preload = function
  | NoPL -> "no"
  | RandomPL -> "random"
  | CustomPL -> "custom"

let set_preload tag = preload := parse_preload tag

let get_preload () = !preload
let default_preload () = pp_preload dpreload

(* Collecting results *)
type collect = CollectLocal|CollectAfter|CollectBoth

let dcollect = CollectAfter
let collect = ref dcollect

let parse_collect = function
  | "local" -> CollectLocal
  | "after" -> CollectAfter
  | "both" -> CollectBoth
  | tag ->  raise (Arg.Bad ("incorrect collect tag: "^tag))

let set_collect tag  = collect := parse_collect tag

let pp_collect = function
  | CollectLocal -> "local"
  | CollectAfter -> "after"
  | CollectBoth -> "both"

let get_collect () = !collect

let default_collect () = pp_collect !collect

(* SMT description *)
type smt_mode = SMTSeq | SMTEnd | SMTNone

let dsmt_mode = SMTNone

let smt_mode = ref dsmt_mode

let parse_smt_mode = function
  | "none" -> SMTNone
  | "seq" -> SMTSeq
  | "end" -> SMTEnd
  | tag ->  raise (Arg.Bad ("incorrect smt_mode tag: "^tag))

let set_smt_mode tag =  smt_mode := parse_smt_mode tag

let pp_smt_mode = function
  |  SMTNone -> "none"
  |  SMTSeq -> "seq"
  |  SMTEnd -> "end"

let get_smt_mode () = !smt_mode
let default_smt_mode () = pp_smt_mode dsmt_mode

let dsmt = 2

let smt = ref dsmt

let set_smt x = smt := x

let get_smt () = !smt

let default_smt () = dsmt

(* Flood avail procs *)
let is_flood arch = match (get_config arch).avail with
| None -> false
| Some _ -> true

let get_avail arch = (get_config arch).avail
and set_avail n = replace_config
    (fun cfg ->
      if n >= 0 then { cfg with avail=Some n }
      else { cfg with avail=None })

(* Technique for parallel testing on big machines *)
type para = Shell | Self

let dp = Self
let para = ref dp

let set_para tag =
  let p = match tag with
  | "self" -> Self
  | "shell" -> Shell
  | _ ->  raise (Arg.Bad ("incorrect para tag: "^tag)) in
  para := p

let get_para () = !para

let do_pp_para = function
  | Self -> "self"
  | Shell -> "shell"

let pp_para () = do_pp_para !para
let default_para () = do_pp_para dp

(* Thread cache *)
let dcache = false
let cache = ref dcache
let set_cache b = cache := b
let get_cache () = !cache
let default_cache () = dcache

(* Call *)
let dcall = false
let call = ref dcall
let set_call b = call := b
let get_call () = !call
let default_call () = dcall

(* Call *)
let dprealloc = false
let prealloc = ref dprealloc
let set_prealloc b = prealloc := b
let get_prealloc () = !prealloc
let default_prealloc () = dprealloc

(* Do not produce tests with more than <limit> threads *)

let limit = ref false
let set_limit b = limit := b
and get_limit () = !limit

(* Avoid cycles *)
let no = ref None
let set_no fname = no := Some fname
and get_no () = !no

(* Hints *)
let hint = ref None
let set_hint fname = hint := Some fname
and get_hint () = !hint

(* Affinity control *)

type affinity =
  | AffinityNone
  | AffinityIncr of int
  | AffinityRandom
  | AffinityCustom

let incr_pref = "incr"
let pref_len = String.length incr_pref

let daffi = AffinityNone
let affi = ref daffi

let get_affinity () = !affi
let set_affinity tag =
  let a = match tag with
  | "none" -> AffinityNone
  | "random"|"rand" -> AffinityRandom
  | "custom" -> AffinityCustom
  | _ ->
      try
        let len = String.length tag in
        if len > pref_len then
          let pref =  String.sub tag 0 pref_len in
          if pref <> incr_pref then raise (Failure incr_pref) ;
          let suff = String.sub tag pref_len (len-pref_len) in
          AffinityIncr (int_of_string suff)
        else          
          raise (Arg.Bad ("incorrect affinity tag: "^tag))
      with Failure _ -> raise (Arg.Bad ("incorrect affinity tag: "^tag)) in
  affi := a

let set_affinity_incr i = affi := AffinityIncr i
  

let do_pp_affi = function
  | AffinityNone -> "none"
  | AffinityIncr i -> sprintf "incr%i" i
  | AffinityRandom -> "random"
  | AffinityCustom -> "custom"

let pp_affinity = do_pp_affi

let default_affinity () = do_pp_affi daffi


(* Logical processor order *)

let logical_procs = ref None

let set_logical_procs s =
  try
    logical_procs := Some (LexSplit.ints s) ;
  with LexSplit.Error ->
    raise (Arg.Bad ("bad logical processors mapping: " ^s))

let get_logical_procs () = !logical_procs

let default_force_affinity = false
let fa = ref default_force_affinity
let set_force_affinity b = fa := b
let get_force_affinity () = !fa


(* bench file (defines number of iterations) *)
let bench = ref None
let set_bench s = bench := Some s
let get_bench () = !bench


(* Index file *)

let exp = ref None

let set_exp s = exp := Some s

let get_exp () = !exp

(*********************************)
(* Insert isync just before test *)
(*********************************)
let disync = false
let isync = ref disync

let default_isync () = disync
let get_isync () = !isync
let set_isync b = isync := b

(***********************)
(* PPC Sync Macro      *)
(***********************)

let sync_const = 128

let sync_macro_number_times = ref None

let set_sync_macro n =  sync_macro_number_times:= Some n
let get_sync_macro () = !sync_macro_number_times


(*****************************)
(* Signal condition directly *)
(*****************************)

let default_signaling_write () = false

let signaling_write = ref (default_signaling_write ())

let get_signaling_write () = !signaling_write
let set_signaling_write b = signaling_write := b

(***************************************)
(* Suppress final condition quantifier *)
(***************************************)

let default_kind () = true

let kind = ref (default_kind ())

let get_kind () = !kind
let set_kind b = kind := b

(**********************)
(* Add sleep commands *)
(**********************)

let default_sleep () = -1

let sleep = ref (default_sleep ())

let get_sleep () = !sleep
let set_sleep n = sleep := n

(*******)
(* gcc *)
(*******)

let default_gcc () = "gcc"

let gcc = ref (default_gcc ())

let get_gcc () = !gcc
and set_gcc s = gcc := s

(***************)
(* link option *)
(***************)

let link_opt = ref ""
let get_link_opt () = !link_opt
and set_link_opt s = link_opt := s

(***********************)
(* Timing data control *)
(***********************)

(**
 * Brief description of issues in timing litmus runs:
 *
 * Litmus tests are done in blocks of (default) 200 runs, times
 * restarting the threads (default) 1000 times. Each block is attempted
 * with as less synchronization as possible, while the restarts take
 * care of (hopefully) resynchronization.
 * 
 * Method 1: Use processor-local Time Stamp Counter (TSC)
 * ------------------------------------------------------
 * This is an extremely high-res counter (measures in CPU cycles) of time.
 * Very processor-dependent, on X86, there is the RDTSC instruction. On
 * PowerPC, there is an untested sequence of instructions with similar 
 * effect.
 * 
 * This does not do any synchronization, and, (at least on X86) is 
 * out-of-order, thus, rdtsc can return even when preceding stores (eg) have
 * not completed. This makes it less useful as a precise timing, but has
 * value in looking at individual tests (even within blocks).
 * 
 * Note that TSC is not guaranteed to be unique across different cores, so
 * values are only meaningful if on same processor. If there is an intervening
 * preemption by scheduler, all bets are off. With that caveat, it should
 * monotonically increase in time (ignoring 64-bit wraparound issues).
 * 
 * Method 2: Use the POSIX function clock_gettime.
 * -----------------------------------------------
 * This is a POSIX function, implemented within Linux [Note1] in
 * all modern systems. The CLOCK_MONOTONIC is guaranteed to return
 * monotonically increasing values, regardless of calling processor.
 * The cost of this is a system call. On tested systems, it has nanosecond 
 * resolution.
 * 
 * Because of the syscall, this has too much synchronization to be 
 * done within blocks. We perform these calls at the beginning and end of
 * each block, per thread. Thus information (at nanosecond resolution) is
 * made available at the block level.
 * 
 * Note that this requires -lrt (real-time library?) to gcc
 * 
 * [Note1]: Googling suggests AIX supports it as well, untested. MacOSX 
 * DOES NOT, apparently. See here for a possible workaround:
 * http://www.wand.net.nz/~smr26/wordpress/2009/01/19/monotonic-time-in-mac-os-x/ 
 *)
let do_tsc = ref false
let get_tsc () = !do_tsc
let set_tsc b = do_tsc := b

let do_clocktime = ref false
let get_clocktime () = !do_clocktime
let set_clocktime b = do_clocktime := b

(* Profiling with machine time stamps *)
let do_profiling = ref false
let get_profiling () = !do_profiling
let set_profiling b = do_profiling := b

(* Alternative: put assembly code in a loop *)
let do_time_loop = ref 0
let get_time_loop () = !do_time_loop
let set_time_loop b = do_time_loop := b

(*******************)
(* Speedcheck mode *)
(*******************)
type speedcheck = NoSpeed | SomeSpeed | AllSpeed

let ds = NoSpeed
let speedcheck = ref ds

let parse_speedcheck tag = match tag with
| "false"|"no" -> NoSpeed
| "true"|"some" -> SomeSpeed
| "all"|"dont" -> AllSpeed
| _ -> raise (Arg.Bad ("incorrect speedcheck tag: "^tag))

let set_speedcheck b = speedcheck := (parse_speedcheck b)
let get_speedcheck () = !speedcheck
let default_speedcheck () = ds

let pp_speedcheck = function
  | NoSpeed -> "no"
  | SomeSpeed -> "some"
  | AllSpeed -> "all"
(************)
(* run name *)
(************)

let runname = ref "default"
let set_runname s = runname := s
let get_runname () = !runname

(*********************)
(* Cross compilation *)
(*********************)

let tar = ref None
let cross = ref false
let cross_mode () = !cross

let set_tar b  = cross := false ; tar := Some b
let set_cross b  = cross := true ; tar := Some b

and is_out () = match !tar with
| Some _ -> true
| None -> false
and get_tar () = match !tar with
| Some s -> s
| None -> assert false

(*************)
(* Cross run *)
(*************)

type addr ={ host : string ; port : int option ; }

type remote =
  | RemoteHost of addr
  | RemoteAdb
  | RemoteNone

let default_crossrun () = RemoteNone
let cr = ref (default_crossrun ())
let get_crossrun () = !cr

let parse_addr s = match s with
| "none" -> RemoteNone
| "adb"  -> RemoteAdb
| _ ->
    let h =
      try
        let j = try String.index s ':' with Not_found -> raise Exit in
        let h = String.sub s 0 j
        and p = String.sub s (j+1) (String.length s - (j+1)) in
        let p = try int_of_string p with _ -> raise Exit in
        { host=h; port=Some p;}
      with
      | Exit -> { host=s ; port=None; } in
    RemoteHost h

let set_crossrun host = cr := parse_addr host

