(*********************************************************************)
(*                        Memevents/Litmus                           *)
(*                                                                   *)
(* Jade Alglave, Luc Maranget, INRIA Paris-Rocquencourt, France.     *)
(* Susmit Sarkar, Peter Sewell, University of Cambridge, UK.         *)
(*                                                                   *)
(*  Copyright 2012 Institut National de Recherche en Informatique et *)
(*  en Automatique and the authors. All rights reserved.             *)
(*  This file is distributed  under the terms of the Lesser GNU      *)
(*  General Public License.                                          *)
(*********************************************************************)

(* Extract name: cycle from test(s) *)

open Printf

let verbose = true

type cycle = One of string | Norm of string * string


module Make
    (A:Arch.S) 
    (L:GenParser.LexParse with type instruction = A.pseudo) =
  struct
    module P = GenParser.Make(A)(L)


    let get_cycle info =
      let orig =
        try Some (List.assoc "Orig" info)
        with Not_found -> None
      and cycle = 
        try Some (List.assoc "Cycle" info)
        with Not_found -> None in
      match orig,cycle with
      | Some o,Some c -> Norm (o,c)
      | None,Some c
      | Some c,None -> One c
      | None,None -> raise Not_found

    let cycle chan splitted =
      let name = splitted.Splitter.name in
      let name = name.Name.name in
      let parsed = P.parse chan splitted in
      name,get_cycle  parsed.MiscParser.info
  end

module G(V:Value.S) = struct
  let from_file name chan = 
  let (splitted:Splitter.result) =  Splitter.split name chan in
  match splitted.Splitter.arch with
  | Misc.PPC ->
      let module PPC = PPCArch.Make(V) in
      let module PPCLexParse = struct
	type instruction = PPC.pseudo
	type token = PPCParser.token

	let lexer = PPCLexer.token
	let parser = PPCParser.main
      end in
      let module X = Make (PPC) (PPCLexParse) in
      X.cycle chan splitted
  | Misc.X86 ->
      let module X86 = X86Arch.Make(V) in
      let module X86LexParse = struct
	type instruction = X86.pseudo
	type token = X86Parser.token

	let lexer = X86Lexer.token
	let parser = X86Parser.main
      end in
      let module X = Make (X86) (X86LexParse) in
      X.cycle chan splitted
  | Misc.ARM ->
      let module ARM = ARMArch.Make(V) in
      let module ARMLexParse = struct
	type instruction = ARM.pseudo
	type token = ARMParser.token

	let lexer = ARMLexer.token
	let parser = ARMParser.main
      end in
      let module X = Make (ARM) (ARMLexParse) in
      X.cycle chan splitted

let from_file name =
  Misc.input_protect
    (from_file name)
    name

let arg = ref []

let prog =
  if Array.length Sys.argv > 0 then Sys.argv.(0)
  else "msort"

let () =
  Arg.parse
    []
    (fun s -> arg := s :: !arg)
    (sprintf "Usage: %s [test]*" prog)

let tests = !arg

module Strings = MySet.Make(String)
let seen = ref Strings.empty
let () =
  Misc.iter_argv
    (fun name ->
      try
        let tname,cycle = from_file name in
        match cycle with
        | One c ->
            printf "%s: %s\n" tname c
        | Norm (o,c) ->
            if not (Strings.mem c !seen) then begin
              seen := Strings.add c !seen ;
              printf "%s: %s\n" tname o
            end
      with
      | Not_found -> Warn.warn_always "%s: no cycle" name
      | Misc.Exit -> ()
      | Misc.Fatal msg ->
          Warn.warn_always "%a %s" Pos.pp_pos0 name msg
      | e ->
	  Printf.eprintf "\nFatal: %a Adios\n" Pos.pp_pos0 name ;
	  raise e)
    tests

end

